#include <ssslib_cpp.h>

comm_lib::comm_lib(void){
  sss = SSS_Init(NULL,0);
}

comm_lib::comm_lib(int debug){
  sss = SSS_Init(NULL,debug);
}

comm_lib::comm_lib(char *conffile){
  sss = SSS_Init(conffile,0);
}

comm_lib::comm_lib(char *conffile,int debug){
  sss = SSS_Init(conffile,debug);
}

comm_lib::~comm_lib(void){
  SSS_Close(sss);
}

int comm_lib::SDRegister(char *service,char *host,int port,char *protocol,char *schema_version,char *tier){
  int rc;

  rc = SSS_SDRegister(sss,service,host,port,protocol,schema_version,tier);

  if (rc == -1){
    throw_exception(sss->sss_errno);
  }

  return rc;
}

int comm_lib::SDRemove(char *service,char *host,int port,char *protocol){
  int rc;

  rc = SSS_SDRemove(sss,service,host,port,protocol);

  if (rc == -1){
    throw_exception(sss->sss_errno);
  }

  return rc;
}

char *comm_lib::ServerInit(int socket){
  char *retval;

  retval = SSS_ServerInit(sss,socket);

  if (!retval){
    throw_exception(sss->sss_errno);
  }

  return retval;
}

char *comm_lib::ClientInit(char *service){
  char *retval;

  retval = SSS_ClientInit(sss,service);

  if (!retval){
    throw_exception(sss->sss_errno);
  }

  return retval;

}

int comm_lib::ClientClose(char *handle){
  int rc;

  rc = SSS_ClientClose(sss,handle);

  if (rc == -1){
    throw_exception(sss->sss_errno);
  }

  return rc;
}

int comm_lib::ServerClose(char *handle){
  int rc;

  rc = SSS_ServerClose(sss,handle);

  if (rc == -1){
    throw_exception(sss->sss_errno);
  }

  return rc;
}

int comm_lib::SendMessage(char *handle,char *message){
  int rc;

  rc = SSS_SendMessage(sss,handle,message);

  if (rc == -1){
    throw_exception(sss->sss_errno);
  }

  return rc;
}

char *comm_lib::RecvMessage(char *handle){
  char *retval;

  retval = SSS_RecvMessage(sss,handle);

  if (!retval){
    throw_exception(sss->sss_errno);
  }

  return retval;
}

void comm_lib::throw_exception(int errnum){
  int typenum;
  char buf[MAXERRLENGTH];

  SSS_strerror(errnum,buf,&typenum);

  switch (typenum) {
  case SSS_PARSE_ERROR:
    throw SSSParseException(buf);
    break;
  case SSS_INIT_ERROR:
    throw SSSInitException(buf);
    break;
  case SSS_PROTOCOL_ERROR:
    throw SSSProtocolException(buf);
    break;
  case SSS_CONNECT_ERROR:
    throw SSSConnectException(buf);
    break;
  case SSS_COMMUNICATION_ERROR:
    throw SSSCommunicationException(buf);
    break;
  case SSS_INTERNAL_ERROR:
    throw SSSInternalException(buf);
    break;
  }
}

SSS_Config_Dictionary *comm_lib::GetConfig(void){
  return new SSS_Config_Dictionary(SSS_GetGlobalConfigDict(sss));
}

