#include<ssslib.h>
#include<stdlib.h>
#include<string.h>
#include<unistd.h>

#include<errno.h>
#include<netinet/in.h>
#include<sys/socket.h>
#include<netdb.h>
#include<openssl/ssl.h>

char *keypw;

struct transport_ssl_state {
  int sss_errno;
  int debug;
  int socket;
  SSL *ssl;
  SSL_CTX *ssl_ctx;
  char *key_file;
  char *cert_file;
};

int get_password(char *buf,int size,int rwflag,void *userdata){
  strncpy(buf,keypw,size);
  return strlen(keypw);
}

void *transport_ssl_LTX_transport_getstate(SSS_ConfigDict cfg,int debug){
  struct transport_ssl_state *ret;

  if (debug)
    printf("transport_ssl: starting init\n");

  ret = malloc(sizeof(struct transport_ssl_state));
  ret->debug = debug;
  ret->ssl = NULL;
  ret->ssl_ctx = NULL;
  ret->key_file = SSS_GetFromConfigDict(cfg,"key_file");
  if (!ret->key_file){
    if (debug)
      fprintf(stderr,"transport_ssl: need a \"key_file\" in configuration file.\n");
    return NULL;
  }
  ret->cert_file = SSS_GetFromConfigDict(cfg,"cert_file");
  if (!ret->cert_file){
    if (debug)
      fprintf(stderr,"transport_ssl: need a \"cert_file\" in configuration file.\n");
    return NULL;
  }
  if (debug)
    printf("transport_ssl: successful init\n");
  keypw = SSS_GetFromConfigDict(cfg,"key_password");
  if (!keypw){
    if (debug)
      fprintf(stderr,"transport_ssl: need a \"key_password\" in configuration file.\n");
    return NULL;
  }

  return ret;
}

int transport_ssl_LTX_serv_connect(void *transport_state,int extra){
  int rc;
  struct transport_ssl_state *state = transport_state;
  if (state->debug)
    printf("transport_ssl: starting serv_connect\n");

  state->socket = extra;
  SSL_library_init();
  SSL_load_error_strings();
  state->ssl_ctx = SSL_CTX_new(SSLv3_server_method());
  if (!state->ssl_ctx){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_CTX_new\n");
    return -1;
  }
  SSL_CTX_set_default_passwd_cb(state->ssl_ctx,get_password);
  rc = SSL_CTX_use_certificate_file(state->ssl_ctx,state->cert_file,SSL_FILETYPE_PEM);
  if (rc != 1){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_CTX_use_certificate_file\n");
    return -1;
  }
  rc = SSL_CTX_use_PrivateKey_file(state->ssl_ctx,state->key_file,SSL_FILETYPE_PEM);
  if (rc != 1){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_CTX_use_PrivateKey_file\n");
    return -1;
  }
  state->ssl = SSL_new(state->ssl_ctx);
  if (!state->ssl){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_new\n");
    return -1;
  }
  rc = SSL_set_fd(state->ssl,state->socket);
  if (rc == 0){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_set_fd\n");
    return -1;
  }
  rc = SSL_accept(state->ssl);
  if (rc < 1){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_accept\n");
    return -1;
  }
  if (state->debug)
    printf("transport_ssl: successful serv_connect\n");
  return 0;
}

int transport_ssl_LTX_cli_connect(void *transport_state,char *host,char *extra){
  int rc;
  struct addrinfo hints;
  struct addrinfo *res;
  struct addrinfo *itr;
  struct transport_ssl_state *state = transport_state;
  
  if (state->debug)
    printf("transport_ssl: starting cli_connect\n");
  
  bzero(&hints,sizeof(struct addrinfo));
  hints.ai_socktype = SOCK_STREAM;
  rc = getaddrinfo(host,extra,&hints,&res);
  if (rc != 0){
    switch(rc) {
    case EAI_NONAME:
      state->sss_errno = SSS_HOST_NOT_FOUND;
      break;
    default:
      state->sss_errno = SSS_UNRECOVERABLE_DNS_ERROR;
      break;
    }
    if (state->debug)
      printf("transport_ssl: cli_connect failed\n");
    return -1;
  }

  itr = res;

  do{

    state->socket = socket(itr->ai_family,SOCK_STREAM,IPPROTO_TCP);
    if (state->socket == -1){
      switch(errno){
      case EPROTONOSUPPORT:
        state->sss_errno = SSS_PROTO_NO_SUPPORT;
        break;
      case EAFNOSUPPORT:
        state->sss_errno = SSS_FAM_NO_SUPPORT;
        break;
      case ENFILE:
      case ENOBUFS:
      case ENOMEM:
        state->sss_errno = SSS_NOMEM;
        break;
      case EACCES:
        state->sss_errno = SSS_NOSOCKPERM;
        break;
      case EINVAL:
        state->sss_errno = SSS_INVALPROT;
        break;
      default:
        if (state->debug)
          printf("transport_ssl: cli_connect socket call unhandled error\n");
        state->sss_errno = SSS_UNKNOWN_ERROR;
      }
      if (state->debug)
        printf("transport_ssl: cli_connect failed\n");
      if (!itr->ai_next){
        freeaddrinfo(res);
        return -1;
      }
    }

  preconnect:
    rc = connect(state->socket,itr->ai_addr,itr->ai_addrlen);
  
    if (rc == -1) {
      switch (errno){
      case EBADF:
        state->sss_errno = SSS_BAD_FD;
        break;
      case EFAULT:
        state->sss_errno = SSS_FAULT;
        break;
      case ENOTSOCK:
        state->sss_errno = SSS_BAD_FD;
        break;
      case EISCONN:
        state->sss_errno = SSS_ALREADY_CONNECTED;
        break;
      case ECONNREFUSED:
        state->sss_errno = SSS_CONN_REFUSED;
        break;
      case ETIMEDOUT:
        state->sss_errno = SSS_TIMEOUT;
        break;
      case ENETUNREACH:
        state->sss_errno = SSS_NET_UNREACH;
        break;
      case EHOSTUNREACH:
        state->sss_errno = SSS_NO_ROUTE;
        break;
      case EADDRINUSE:
        state->sss_errno = SSS_ADDR_IN_USE;
        break;
      case EAGAIN:
        state->sss_errno = SSS_NO_FREE_PORTS;
        break;
      case ECONNRESET:
        state->sss_errno = SSS_CONN_RESET;
        break;
      case EACCES:
      case EPERM:
        state->sss_errno = SSS_NO_CONNECT_PERM;
        break;
      case EINTR:
        goto preconnect;
        break;
      default:
        if (state->debug)
          printf("transport_ssl: cli_connect connect call unhandled error\n");
        state->sss_errno = SSS_UNKNOWN_ERROR;
      }
      if (state->debug)
        printf("transport_ssl: cli_connect failed\n");
      if (!itr->ai_next){
        freeaddrinfo(res);
        return -1;
      }
      close(state->socket);
    }

    itr = itr->ai_next;

  } while (itr);
  
  freeaddrinfo(res);

  SSL_library_init();
  SSL_load_error_strings();
  state->ssl_ctx = SSL_CTX_new(SSLv3_client_method());
  if (!state->ssl_ctx){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_CTX_new\n");
    return -1;
  }
  state->ssl = SSL_new(state->ssl_ctx);
  if (!state->ssl){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_new\n");
    return -1;
  }
  rc = SSL_set_fd(state->ssl,state->socket);
  if (rc == 0){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_set_fd\n");
    return -1;
  }
  rc = SSL_connect(state->ssl);
  if (rc < 1){
    state->sss_errno = SSS_PROT_ERROR;
    if (state->debug)
      fprintf(stderr,"transport_ssl: error in SSL_connect\n");
    return -1;
  }
  if (state->debug)
    printf("transport_ssl: successful cli_connect\n");

  return 0;
}

int transport_ssl_LTX_transport_write(void *transport_state,void *buf,int count){
  int ssl_rc;
  int rc;

  struct transport_ssl_state *state = transport_state;

  if (state->debug)
    printf("transport_ssl: starting write\n");

  rc = SSL_write(state->ssl,buf,count);
  if (rc <= 0){
    ssl_rc = SSL_get_error(state->ssl,rc);
    if (ssl_rc == SSL_ERROR_SYSCALL){
      switch(errno){
      case EBADF:
        state->sss_errno = SSS_BAD_FD;
        break;
      case ENOTSOCK:
        state->sss_errno = SSS_NOT_A_SOCKET;
        break;
      case EFAULT:
        state->sss_errno = SSS_FAULT;
        break;
      case EMSGSIZE:
        state->sss_errno = SSS_INVALID_SIZE;
        break;
      case ENOBUFS:
        state->sss_errno = SSS_NOMEM;
        break;
      case EINTR:
        state->sss_errno = SSS_INTERRUPT;
        break;
      case ENOMEM:
        state->sss_errno = SSS_NOMEM;
        break;
      case EINVAL:
        state->sss_errno = SSS_INVAL_ARG;
        break;
      case EPIPE:
        state->sss_errno = SSS_BROKEN_PIPE;
        break;
      case ECONNRESET:
        state->sss_errno = SSS_CONN_RESET;
        break;
      default:
        if (state->debug)
          printf("transport_ssl: write call unhandled error, errno is %i, ssl_rc is %i, rc is %i\n",errno,ssl_rc,rc);
        state->sss_errno = SSS_UNKNOWN_ERROR;
      }
      return -1;
    }
    else{
      state->sss_errno = SSS_PROT_ERROR;
      if (state->debug)
        fprintf(stderr,"transport_ssl: error in SSL_write\n");
      return -1;
    }
  }

  if (state->debug)
    printf("transport_ssl: successful write\n");
  
  return rc;
}

int transport_ssl_LTX_transport_writen(void *transport_state,char *buf,int count){
  int nleft, nwritten;
  struct transport_ssl_state *state = transport_state;

  if (state->debug)
    printf("transport_ssl: starting writen\n");

  nleft = count;
  while (nleft > 0)   {
    nwritten = transport_ssl_LTX_transport_write(transport_state, buf, nleft);
    if (nwritten <= 0)
      return -1;
    nleft -= nwritten;
    buf += nwritten;
  }
  if (state->debug)
    printf("transport_ssl: successful writen\n");
  return 0;
}

int transport_ssl_LTX_transport_read(void *transport_state,char *buf,int count){
  int ssl_rc;
  int rc;
  struct transport_ssl_state *state = transport_state;

  if (state->debug)
    printf("transport_ssl: starting read\n");

  rc = SSL_read(state->ssl,buf,count);
  if (rc <= 0){
    ssl_rc = SSL_get_error(state->ssl,rc);
    if (ssl_rc == SSL_ERROR_SYSCALL){
      switch(errno){
      case EBADF:
        state->sss_errno = SSS_BAD_FD;
        break;
      case ECONNREFUSED:
        state->sss_errno = SSS_CONN_REFUSED;
        break;
      case ENOTCONN:
        state->sss_errno = SSS_SOCK_NOT_CONNECTED;
        break;
      case EINTR:
        state->sss_errno = SSS_INTERRUPT;
        break;
      case EFAULT:
        state->sss_errno = SSS_FAULT;
        break;
      case EINVAL:
        state->sss_errno = SSS_INVAL_ARG;
        break;
      case ECONNRESET:
        state->sss_errno = SSS_CONN_RESET;
        break;
      case ESPIPE:
        state->sss_errno = SSS_ILLEGAL_SEEK;
        break;
      default:
        if (rc == 0){
          state->sss_errno = SSS_EOF;
          return -1;
        }
        if (state->debug)
          printf("transport_ssl: read call unhandled error, errno is %i, ssl_rc is %i, rc is %i\n",errno,ssl_rc,rc);
        state->sss_errno = SSS_UNKNOWN_ERROR;
      }
      return -1;
    }
    else if (ssl_rc == SSL_ERROR_ZERO_RETURN){
      state->sss_errno = SSS_EOF;
      return -1;
    }
    else{
      state->sss_errno = SSS_PROT_ERROR;
      if(state->debug)
        fprintf(stderr,"transport_ssl: error in SSL_read\n");
      return -1;
    }
  }

  if (state->debug)
    printf("transport_ssl: successful read\n");
  
  return rc;
}

int transport_ssl_LTX_transport_readn(void *transport_state,char *buf,int count){
  int nleft, nread;
  struct transport_ssl_state *state = transport_state;
  
  if (state->debug)
    printf("transport_ssl: starting readn\n");

  nleft = count;
  while (nleft > 0){
    nread = transport_ssl_LTX_transport_read(transport_state, buf, nleft);
    if (nread < 0)
      return nread;
    else if (nread == 0)
      break;
    nleft -= nread;
    buf += nread;
  }

  if (state->debug)
    printf("transport_ssl: successful readn\n");

  return 0;

}

int transport_ssl_LTX_transport_close(void *transport_state){
  int debug;
  struct transport_ssl_state *state = transport_state;
  
  if (state->debug)
    printf("transport_ssl: starting transport_close\n");

  close(state->socket);
  debug = state->debug;
    
  free(state);

  if (debug)
    printf("transport_ssl: successful transport_close\n");

  return 0;
}
