#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <linux/if.h>
#include <netinet/in.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

struct ifmapping {
  char *iface;
  char *hwaddr;
  char *driver;
};

char *read_line(int);
char *get_iface(int);
int is_in_list(struct ifmapping **,char *);
struct ifmapping **testnic(char *,struct ifmapping **,int);
void identify_serial(int,char *,struct ifmapping **);
void identify_random(int,char *,struct ifmapping **);
void identify_snmp(int,char *,struct ifmapping **);

int main(int argc, char *argv[]){

  int i;
  struct ifmapping **ifs = malloc(sizeof(struct ifmapping *));
  int idserv_sock;
  struct hostent *ipaddr;
  struct sockaddr_in addr;
  char *readbuf;
  char *token;
  char *ident_type;
  int rc;
  int pid;
  int status;
  int exec;
  ifs[0] = NULL;

  if (argc < 5){
    fprintf(stderr,"Usage: ifs <Identify server> <port> <dhcp(1 or 0> <driver1> [driver2] [driver3] ...\n");
    exit(1);
  }

  exec = atoi(argv[3]);

  for (i=4;i<argc;i++)
    ifs = testnic(argv[i],ifs,exec);

  fprintf(stdout, "About to call dhcpcd\n");
  fflush(stdout);

  if (exec){
  pid = fork();

  if (!pid){
    rc = execl("/bin/dhcpcd","/bin/dhcpcd",NULL);
    printf("problem with exec!\n");
    perror("");
  }
  
  waitpid(pid,&status,0);

  if (WEXITSTATUS(status)){
    fprintf(stderr,"dhcpcd met with failure\n");
    exit(1);
  }
  
  printf("finished with dhcpcd\n");
  }
  rc = idserv_sock = socket(AF_INET,SOCK_STREAM,0);
  if (rc == -1){
    perror("");
    exit(1);
  }
  ipaddr = gethostbyname(argv[1]);
  if (!ipaddr){
    perror("");
    exit(1);
  }
  addr.sin_family = AF_INET;
  addr.sin_port = htons(atoi(argv[2]));
  memcpy(&(addr.sin_addr.s_addr),*(ipaddr->h_addr_list),sizeof(struct in_addr));
  memset(&(addr.sin_zero),'\0',8);
  rc = connect(idserv_sock,(struct sockaddr *)&addr,sizeof(struct sockaddr));
  if (rc == -1){
    perror("");
    exit(2);
  }
  send(idserv_sock,"identify\n",9,0);
  send(idserv_sock,"node\n",5,0);
  readbuf=read_line(idserv_sock);
  readbuf[strlen(readbuf)-1] = '\0';
  printf("readbuf back from idserv is :%s:\n",readbuf);
  token = strdup(strstr(readbuf," ")+1);
  readbuf[strstr(readbuf," ") - readbuf] = '\0';
  ident_type = strdup(readbuf);
  free(readbuf);

  if (!strcmp(ident_type,"ident_serial_console"))
      identify_serial(idserv_sock,token,ifs);
  else if (!strcmp(ident_type,"ident_random"))
      identify_random(idserv_sock,token,ifs);
  else if (!strcmp(ident_type,"ident_snmp"))
      identify_snmp(idserv_sock,token,ifs);

  free(ident_type);
  
  return 0;
}

struct ifmapping **testnic(char *driver_to_test,struct ifmapping **have,int exec){
  int sock;
  int rc;
  unsigned char *hwaddr;
  int proc_net_dev;
  char *ifname;
  struct ifreq ifr;
  char hwbuf[18];
  char *sysbuf;
  struct ifmapping *temp;
  int j;
  int status;
  int pid;

  if (exec){
  sysbuf = malloc(snprintf(NULL,0,"insmod %s",driver_to_test)+1);

  printf("about to insmod\n");
  fflush(stdout);
  
  sprintf(sysbuf,"insmod %s",driver_to_test);

  pid = fork();

  if (!pid)
    execl("/bin/insmod","/bin/insmod",driver_to_test,NULL);
  
  waitpid(pid,&status,0);
  }
  proc_net_dev = open("/proc/net/dev",0);

  free(read_line(proc_net_dev));
  free(read_line(proc_net_dev));

  sock = socket(AF_INET,SOCK_STREAM,IPPROTO_IP);
  
  while (ifname = get_iface(proc_net_dev)){
    if (is_in_list(have,ifname) || !strcmp(ifname,"lo"))
      continue;
    strncpy(ifr.ifr_ifrn.ifrn_name,ifname,IFNAMSIZ-1);
    rc = ioctl(sock,SIOCGIFHWADDR,&ifr);
    hwaddr = (unsigned char *)ifr.ifr_hwaddr.sa_data;
    sprintf(hwbuf,"%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x",hwaddr[0],
	   hwaddr[1],hwaddr[2],hwaddr[3],hwaddr[4],hwaddr[5]);
    temp = malloc(sizeof(struct ifmapping));
    temp->iface = ifname;
    temp->hwaddr = strdup(hwbuf);
    temp->driver = strdup(driver_to_test);
    for(j=0;have[j];j++){}
    have = realloc(have,(j+2)*sizeof(struct ifmapping *));
    have[j+1] = NULL;
    have[j] = temp;
  }
    
  close(proc_net_dev);

  return have;
}

char *read_line(int stream){
  char small_buf[1];
  int linelen = 1;
  char *retptr = malloc(1);
  
  retptr[0] = '\0';

  small_buf[0] = '\0';

  while(small_buf[0] != '\n'){
    if (!read(stream,small_buf,1))
      return NULL;
    retptr = realloc(retptr,++linelen);
    retptr[linelen-1] = '\0';
    retptr[linelen-2] = small_buf[0];
  }
  return retptr;
}

char *get_iface(int stream){
  char *buf = read_line(stream);
  char *tbuf = buf;
  if (!buf)
    return NULL;
  buf[strstr(buf,":")-buf] = '\0';
  buf = strpbrk(buf,"abcdefghijklmnopqrstuvwxyz");
  buf = strdup(buf);
  free(tbuf);
  return buf;
}

int is_in_list(struct ifmapping **list,char *name){
  int i;
  for(i=0;list[i];i++)
    if (!strcmp(list[i]->iface,name))
      return 1;

  return 0;
}

void identify_serial(int idserv_sock,char *token,struct ifmapping **ifs){
  int i;
  int buflen;
  char *sendbuf;
  int rc;

  for(i=0;i<10;i++)
	printf("\n");
  
  sleep(5);
  
  fprintf(stdout,"IDENTIFY id_token=%s\n",token);
  fflush(stdout);

  for(i=0;i<10;i++)
	  printf("\n");
  
  for (i=0;ifs[i];i++){
    buflen=snprintf(NULL,0,"%s,%s\n",ifs[i]->hwaddr,ifs[i]->driver);
    sendbuf = malloc(buflen+1);
    sprintf(sendbuf,"%s,%s\n",ifs[i]->hwaddr,ifs[i]->driver);
    rc = send(idserv_sock,sendbuf,buflen,0);
    free(sendbuf);
  }
  
}

void identify_random(int idserv_sock,char *token,struct ifmapping **ifs){
  int i;
  int buflen;
  char *sendbuf;
  int rc;

  for (i=0;ifs[i];i++){
    buflen=snprintf(NULL,0,"%s,%s\n",ifs[i]->hwaddr,ifs[i]->driver);
    sendbuf = malloc(buflen+1);
    sprintf(sendbuf,"%s,%s\n",ifs[i]->hwaddr,ifs[i]->driver);
    rc = send(idserv_sock,sendbuf,buflen,0);
    free(sendbuf);
  }

}

void identify_snmp(int idserv_sock,char *token,struct ifmapping **ifs){

}

